<?php
/*
	HMK : Chiffrement par flot. Démonstration online.
	Copyright © 2012 P.Lheureux (lheureux@free.fr) & D.Mestdagh (dimitri.mestdagh@laposte.net)
	
    This file is part of HMK.
    HMK is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published 
	by the Free Software Foundation, either version 3 of the License, 
	or (at your option) any later version.

    HMK is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with HMK. If not, see <http://www.gnu.org/licenses/>.
	
*/

class Hashmask {

	/**
	 * @param $debug TRUE, retire les aléas et laisse valide le chiffré
	 */
	function __construct($debug=FALSE) {
		$this->debug = $debug;
		$this->nomFichierCleOut = 'hmk.key';
		$this->nomFichierCleIn = 'key.dat';
		if(is_file($this->nomFichierOut)){
			unlink($this->nomFichierOut);
		}		
	}
	
	/* Pour l'affichage des erreurs / exceptions : */
	public function afficheErreur($e){
		echo(utf8_decode($e)."<br />\n");
		exit;
	}
	
	/* Création d'un fichier clé avec une clé aléatoire: */
	public function generationCle($nomFichierCle, $longueurCle){		
		
		/* vérifications d'usage : */
		$longueur = $this->verifLongueurCle($longueurCle);
		try {			
			@touch($nomFichierCle);
			if (!is_writable($nomFichierCle)) {
				throw new Exception("Fichier <b>" . $nomFichierCle . "</b> impossible à écrire !");
			}			
		} catch (Exception $e) {
			$this->afficheErreur($e->getMessage());
		}

		/* création d'une clé : */
		for($i=0;$i<($longueurCle/8);$i++){
			$cle .= sprintf("%c",rand(32,126));			
		}
		
		$this->write2file($nomFichierCle,$cle);
				
		$this->getFile($nomFichierCle);
		
		return;
	
	}
		
	/* Création d'un fichier clé avec contenant la clé spécifiée par l'utilisateur : */
	public function sauverCle($nomFichierCle, $cle){	
	
		$this->write2file($nomFichierCle,$cle);
		
		$this->getFile($nomFichierCle);
		
		return;
	
	}
	
	/* Vérification de la longueur de clé demandée : */
	private function verifLongueurCle($longueurCle){
		if($longueurCle<128){
			$this->afficheErreur("Longueur de clé invalide ! ($longueurCle bits)");
		}else{
			if($longueurCle>=512){
				$longueurCle = 512;
			}else if($longueurCle>=256){
				$longueurCle = 256;
			}else if($longueurCle>=128){
				$longueurCle = 128;
			}	
			if($this->debug){
				echo("Longueur de clé : ($longueurCle bits) <br /><br />");
			}				
			return($longueurCle);
		}
	}
	
	/* Récupération de la clé reçue : */
	public function chargerCle($cle){	
		
		/*  vérification de la clé : */
		$this->longueurCle = $this->verifLongueurCle(strlen($cle)*8);
		$this->cle = $cle;
		
		if($this->debug){
			echo("Clé : <br />");
			var_dump($this->cle);
			echo("<br /><br />");
		}
		
	}

	/* Récupération des données issues du fichier clé éventuellement reçu : */
	public function getKeyData($nomFichierCle, $fichier){
	
		if(!move_uploaded_file($_FILES[$fichier]['tmp_name'], "./".$nomFichierCle)){
			/* Si l'upload n'a pas fonctionné : */
			$this->afficheErreur("Echec de l'envoi du fichier clé !");			  
		 }
	
		try {
			if (!is_readable($nomFichierCle)) {
				throw new Exception("Fichier <b>" . $_FILES[$fichier]['name'] . "</b> impossible à lire !");
			}		
		} catch (Exception $e) {
			$this->afficheErreur($e);			
		}
				
		$handle = @fopen($nomFichierCle, "r");
		if ($handle) {
			while (($buffer = fgets($handle, 2048)) !== false) {
				$contenu .= $buffer;
			}
			if (!feof($handle)) {
				$this->afficheErreur("Erreur: échec lors de la lecture du fichier clé !");
			}
			fclose($handle);
		}
		
		unlink($nomFichierCle);
		
		return($contenu);
	
	}	
	
	/* Construction d'un masque du double de la longueur du texte à chiffrer : */
	private function creerMasque($reset=false){
		
		if(!isset($this->hasch) || $reset==true){
			$this->hasch = hash('sha512', $this->cleAleatoire.$this->cle);
		}
		
		for($i=0;$i<64;$i++){
			$this->hasch = hash('sha512', $this->hasch.$this->cle);
			$masque .= $this->hasch;
		}
		$this->nb_masques += 64;
		
		$this->masque = $masque;
		
		/*
		if($this->debug){
			echo("Masque : <br />");
			var_dump($masque);
			echo("<br /><br />");			
		}	
		*/
	}
	
	/* Masquage du texte : */
	private function masquer($chars){
		
		for($i=0;$i<strlen($chars);$i++){
			$j = $i*2;
			$c =  hexdec($this->masque[$j].$this->masque[$j+1]);
			$chars_masques .= chr( ord($chars[$i]) ^ $c );
		}
		
		return($chars_masques);
		
	}	
		
	/* Démasquage du texte : */
	private function demasquer($chars){
		
		for($i=0;$i<strlen($chars);$i++){
			$j = $i*2;
			$c =  hexdec($this->masque[$j].$this->masque[$j+1]);
			$chars_demasques .= chr($c ^ ord($chars[$i]));
		}
		
		return($chars_demasques);
		
	}
	
	/* Fonction pour retourner/inverser les caractères qui composent le nom du fichier : */
	private function envers($nom){
		for($i=strlen($nom);$i>0;$i--){
			$name .= $nom[$i-1];
		}
		return($name);
	}
	
	/* Chiffrement et déchiffrement : */
	public function cypher($nomFichierIn, $chiffrement){
	
		if(!move_uploaded_file($_FILES[$nomFichierIn]['tmp_name'], "./".$nomFichierIn)){
			/* Si l'upload n'a pas fonctionné : */
			$this->afficheErreur("Echec de l'envoi du fichier de données !");			  
		 }
	
		try {
			if (!is_readable($nomFichierIn)) {
				throw new Exception("Fichier <b>" . $_FILES[$nomFichierIn]['name'] . "</b> impossible à lire !");
			}		
		} catch (Exception $e) {
			$this->afficheErreur($e->getMessage());			
		}
		
		if($chiffrement){
			$this->nomFichierOut = $_FILES[$nomFichierIn]['name'].".hmk";			
		}else{
			$this->nomFichierOut = substr($_FILES[$nomFichierIn]['name'], 0, -4);
		}
		if(file_exists("./".$this->nomFichierOut)){
			unlink("./".$this->nomFichierOut);
		}
		if($this->debug){
			echo("Taille du fichier : <br />");
			var_dump(filesize($nomFichierIn));
			echo("<br /><br />");	
			echo("Nom du fichier en sortie : <br />");
			var_dump($this->nomFichierOut);
			echo("<br /><br />");	
		}		
		
		/* On crée le fichier à télécharger : */
		try {			
			@touch($this->nomFichierOut);
			if (!is_writable($this->nomFichierOut)) {
				throw new Exception("Fichier <b>" . $this->nomFichierOut . "</b> impossible à écrire !");
			}			
		} catch (Exception $e) {
			$this->afficheErreur($e->getMessage());
		}
		
		/* on récupère le hash du fichier original : */
		if($chiffrement){
			/* création d'une clé aléatoire : */
			for($i=0;$i<strlen($this->cle);$i++){
				$this->cleAleatoire .= sprintf("%c",rand(32,126));
			}
			$this->fileHasch = hash('sha256',file_get_contents($nomFichierIn).$this->cle);		
			$this->write2file($this->nomFichierOut,$this->cleAleatoire);		
			if($this->debug){
				echo("Hasch du fichier original : <br />");
				var_dump($this->fileHasch);
				echo("<br /><br />");			
			}			
		}else{
			$this->cleAleatoire = substr(file_get_contents($nomFichierIn), 0, strlen($this->cle));
		}
		
		$handle = @fopen($nomFichierIn, "r");
		
		$debut = microtime(); /* vitesse d'AES (à battre ?) : 110 MB /sec */
		
		if ($handle) {
			$segment = 0;
			while (!feof($handle)) {		  	
				$this->creerMasque();
				if($chiffrement){					
					if($segment==0){
						$contenu = fread($handle, 4032);
						$contenu = $this->fileHasch.$contenu;
					}else{
						$contenu = fread($handle, 4096);
					}
					$data = $this->masquer($contenu);
				}else{
					/* on extrait la clé aléatoire depuis le fichier chiffré : */
					if($segment==0){
						$contenu = substr(fread($handle, 4096+strlen($this->cle)), strlen($this->cle));
					}else{
						$contenu = fread($handle, 4096);
					}
					$data = $this->demasquer($contenu);
					if($segment==0){
						/* on extrait le hash du fichier original depuis le fichier chiffré : */
						$this->fileHasch = substr($data, 0, 64);						
						$data = substr($data, 64);
					}				
				}
				/* on place les données traitées dans le fichier téléchargeable : */
				$this->write2file($this->nomFichierOut,$data);
				$segment=1;
			}
			if (!feof($handle)) {
				$this->afficheErreur("Erreur: échec lors de la lecture du fichier de données !");
			}
			fclose($handle);
			
		}
		
		$fin = microtime();
	   
		if($this->debug){	   
			/* on affiche le temps d'exécution (on multiplie par 1000 car on veut des millisecondes pas des micro-sec) : */
			$temps = ($fin - $debut)*1000;
			echo("Temps d'exécution pour le chiffrement/déchiffrement : ".number_format($temps, 2)." ms <br /><br />");
			echo("Nombre de SHA-512 générés : ".$this->nb_masques." hashes. <br /><br />");
			
		}			
		
		unlink($nomFichierIn);
		
		/* en cas de déchiffrement : */
		if(!$chiffrement){
			
			/* on vérifie l'intégrité du fichier : */
			$verifFileHash = hash('sha256',@file_get_contents($this->nomFichierOut).$this->cle);
			if($this->debug){
				echo("Hasch du fichier original : <br />");
				var_dump($this->fileHasch);
				echo("<br /><br />");
				echo("Hasch du fichier généré par le déchiffrement (à vérifier) : <br />");
				var_dump($verifFileHash);
				echo("<br /><br />");			
			}
			if($verifFileHash != $this->fileHasch){
				$erreur = ("Le fichier semble corrompu, ou vous n'utilisez pas la bonne clef !");
				@unlink($this->nomFichierOut);
				$this->afficheErreur($erreur);				
			}

		}
		
		/* envoi du fichier demandé par l'utilisateur : */
		$this->getFile($this->nomFichierOut);
		
		return;
	
	}
	
	/* pour écrire dans le fichier en sortie : */
	private function write2file($file, $data){
				
		$fichier = fopen($file,'a'); 
		fwrite($fichier,$data); 
		fclose($fichier);
		
		return;
	
	}	
	
	private function getFile($nomFichier){
		if(!$this->debug){		
			/* Redirection de la page vers le téléchargement du fichier : */
			header("Content-disposition: attachment; filename=".$nomFichier);
			header("Content-Type: application/force-download");
			header("Content-Transfer-Encoding: text/plain\n");
			header("Content-Length: ".filesize($nomFichier));
			header("Pragma: no-cache");
			header("Cache-Control: must-revalidate, post-check=0, pre-check=0, public");
			header("Expires: 0");
		}else{
			echo("Contenu du fichier en sortie : <br />");
		}
		@readfile($nomFichier);								
		@unlink($nomFichier);
		exit;
	
	}
	
}

?>